package org.sonatype.tycho.p2.publisher;

import static org.codehaus.plexus.util.StringUtils.interpolate;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.codehaus.plexus.util.StringUtils;
import org.sonatype.aether.artifact.Artifact;
import org.sonatype.aether.graph.DependencyFilter;
import org.sonatype.aether.graph.DependencyNode;
import org.sonatype.aether.graph.DependencyVisitor;
import org.sonatype.aether.util.filter.AndDependencyFilter;
import org.sonatype.aether.util.filter.PatternExclusionsDependencyFilter;
import org.sonatype.aether.util.filter.PatternInclusionsDependencyFilter;
import org.sonatype.tycho.p2.publisher.facade.BundlePublisher;
import org.sonatype.tycho.p2.publisher.impl.DefaultArtifactsFilter;
import org.sonatype.tycho.p2.publisher.impl.Utils;
import org.sonatype.tycho.p2.publisher.model.Advice;
import org.sonatype.tycho.p2.publisher.model.GenerateSettings;
import org.sonatype.tycho.p2.publisher.model.InstallableUnit;
import org.sonatype.tycho.p2.publisher.model.InstallableUnitArtifact;
import org.sonatype.tycho.p2.publisher.model.OnOffSwitch;
import org.sonatype.tycho.p2.publisher.model.PatchChange;
import org.sonatype.tycho.p2.publisher.model.PatchChangeRequirement;
import org.sonatype.tycho.p2.publisher.model.PatchScope;
import org.sonatype.tycho.p2.publisher.model.Property;
import org.sonatype.tycho.p2.publisher.model.ProvidedCapability;
import org.sonatype.tycho.p2.publisher.model.RequiredCapability;
import org.sonatype.tycho.p2.publisher.model.TouchpointData;
import org.sonatype.tycho.p2.publisher.model.TouchpointInstruction;
import org.sonatype.tycho.p2.publisher.model.TouchpointType;
import org.sonatype.tycho.p2.publisher.model.UpdateDescriptor;
import org.sonatype.tycho.p2.publisher.model.artifact.InstallableArtifact;

public class PublishAction
{

    private DependencyNode dependencyTree;

    private final Collection<Advice> advices;

    private final Map<IUKey, InstallableUnit> units;

    private final List<InstallableUnit> rootUnits;

    private final Set<IUKey> metaRequirements;

    private final List<InstallableUnit> processableUnits;

    private final List<InstallableUnit> processableUnitsNotApplicableToMetaRequirements;

    private final List<Artifact> artifacts;

    private final List<InstallableArtifact> installableArtifacts;

    private final HashMap<String, String> globalProperties;

    private final BundlePublisher bundlePublisher;

    private final PublisherLifecycleParticipant participant;

    private final String timestamp;

    public PublishAction( final DependencyNode dependencyTree, final PublisherLifecycleParticipant participant,
                          final Properties properties, final BundlePublisher bundlePublisher )
    {
        this.participant = CompositePublisherLifecycleParticipant.composite( new DefaultArtifactsFilter(), participant );
        this.bundlePublisher = bundlePublisher;
        this.dependencyTree = dependencyTree;
        this.globalProperties = new HashMap<String, String>();
        for ( final Map.Entry<Object, Object> entry : properties.entrySet() )
        {
            this.globalProperties.put( "maven:" + (String) entry.getKey(), (String) entry.getValue() );
        }

        advices = new ArrayList<Advice>();
        units = new HashMap<IUKey, InstallableUnit>();
        metaRequirements = new HashSet<IUKey>();
        rootUnits = new ArrayList<InstallableUnit>();
        processableUnits = new ArrayList<InstallableUnit>();
        processableUnitsNotApplicableToMetaRequirements = new ArrayList<InstallableUnit>();
        artifacts = new ArrayList<Artifact>();
        installableArtifacts = new ArrayList<InstallableArtifact>();

        timestamp = String.format( "%1$tY%1$tm%1$td-%1$tH%1$tM%1$tS", System.currentTimeMillis() );
    }

    public PublisherResult execute()
    {
        dependencyTree = participant.processDependencyTree( dependencyTree );
        advices.addAll( participant.getAdvices( dependencyTree ) );
        dependencyTree = participant.cleanupDependencyTree( dependencyTree );

        processArtifacts();
        mergeUnits();
        addRootUnits();

        return new PublisherResult( new ArrayList<InstallableUnit>( units.values() ), rootUnits, installableArtifacts, artifacts );
    }

    private void mergeUnits()
    {
        for ( final InstallableUnit mergeable : processableUnits )
        {
            final InstallableUnit iu = getInstallableUnit( mergeable.getId(), mergeable.getVersion() );
            merge( mergeable, iu );
        }
        for ( final InstallableUnit mergeable : processableUnitsNotApplicableToMetaRequirements )
        {
            final InstallableUnit iu = getInstallableUnit( mergeable.getId(), mergeable.getVersion() );
            if ( !metaRequirements.contains( new IUKey( iu.getId(), iu.getVersion() ) ) )
            {
                merge( mergeable, iu );
            }
        }
    }

    private void addRootUnits()
    {
        final Map<String, String> namespace = createNamespaceMap( dependencyTree.getDependency().getArtifact() );

        final InstallableUnit rootUnit = new InstallableUnit();
        rootUnit.setId( interpolate( "${maven:id}:root", namespace ) );
        rootUnit.setVersion( interpolate( "${maven:buildVersion}", namespace ) );
        rootUnit.setSingleton( true );

        {
            final UpdateDescriptor updateDescriptor = new UpdateDescriptor();
            updateDescriptor.setId( rootUnit.getId() );
            updateDescriptor.setRange( String.format( "[0,%s)", rootUnit.getVersion() ) );
            updateDescriptor.setSeverity( "0" );

            rootUnit.setUpdateDescriptor( updateDescriptor );
        }

        {
            final ProvidedCapability provided = new ProvidedCapability();
            provided.setNamespace( "org.eclipse.equinox.p2.iu" );
            provided.setName( rootUnit.getId() );
            provided.setVersion( rootUnit.getVersion() );

            rootUnit.addProvidedCapability( provided );
        }

        for ( final InstallableUnit unit : units.values() )
        {
            final ProvidedCapability provided = new ProvidedCapability();
            provided.setName( unit.getId() );
            provided.setNamespace( "org.sonatype.runtime.installable" );
            provided.setVersion( unit.getVersion() );
            unit.addProvidedCapability( provided );

            final RequiredCapability required = new RequiredCapability();
            required.setName( provided.getName() );
            required.setNamespace( provided.getNamespace() );

            if ( metaRequirements.contains( new IUKey( unit.getId(), unit.getVersion() ) ) )
            {
                // required.setRange( String.format( "%1$s", provided.getVersion() ) );
                rootUnit.addMetaRequirement( required );
            }
            else
            {
                required.setRange( String.format( "[%1$s,%1$s]", provided.getVersion() ) );
                rootUnit.addRequiredCapability( required );
            }
        }

        units.put( new IUKey( rootUnit.getId(), rootUnit.getVersion() ), rootUnit );
        rootUnits.add( rootUnit );
    }

    private void processArtifacts()
    {
        dependencyTree.accept( new ArtifactsProcessor() );
    }

    private static void merge( final InstallableUnit source, final InstallableUnit target )
    {
        // TODO we need some merge policy
        if ( source.getVersion() != null )
        {
            target.setVersion( source.getVersion() );
            if ( source.getUpdateDescriptor() != null )
            {
                target.setUpdateDescriptor( source.getUpdateDescriptor() );
            }
        }

        for ( final Property property : source.getProperties() )
        {
            target.addProperty( property );
        }

        for ( final ProvidedCapability capability : source.getProvidedCapabilities() )
        {
            target.addProvidedCapability( capability );
        }

        for ( final RequiredCapability capability : source.getRequiredCapabilities() )
        {
            target.addRequiredCapability( capability );
        }

        for ( final RequiredCapability capability : source.getMetaRequirements() )
        {
            target.addMetaRequirement( capability );
        }

        for ( final RequiredCapability capability : source.getHostRequirements() )
        {
            target.addHostRequirement( capability );
        }

        for ( final PatchScope scope : source.getPatchScope() )
        {
            target.addPatchScope( scope );
        }

        for ( final PatchChange change : source.getPatchChanges() )
        {
            target.addPatchChange( change );
        }

        if ( source.getPatchLifeCycle() != null )
        {
            target.setPatchLifeCycle( source.getPatchLifeCycle() );
        }

        for ( final InstallableUnitArtifact artifact : source.getArtifacts() )
        {
            target.addArtifact( artifact );
        }

        if ( source.getTouchpointType() != null )
        {
            target.setTouchpointType( source.getTouchpointType() );
        }

        if ( source.getTouchpointData() != null )
        {
            for ( final TouchpointInstruction sourceInstruction : source.getTouchpointData().getInstructions() )
            {
                if ( target.getTouchpointData() == null )
                {
                    target.setTouchpointData( new TouchpointData() );
                }

                if ( target.getTouchpointData().getInstructions() != null )
                {
                    boolean found = false;
                    for ( final TouchpointInstruction targetInstruction : target.getTouchpointData().getInstructions() )
                    {
                        if ( sourceInstruction.getKey().equals( targetInstruction.getKey() ) )
                        {
                            // TODO does it need an ; between?
                            targetInstruction.setBody( targetInstruction.getBody() + sourceInstruction.getBody() );
                            found = true;
                            break;
                        }
                    }
                    if ( !found )
                    {
                        target.getTouchpointData().addInstruction( sourceInstruction );
                    }
                }
                else
                {
                    target.getTouchpointData().addInstruction( sourceInstruction );
                }
            }
        }
    }

    private Map<String, String> createNamespaceMap( final Artifact artifact )
    {
        final Map<String, String> namespace = new HashMap<String, String>();

        namespace.putAll( globalProperties );

        namespace.put( "maven:id", Utils.conflictArtifactId( artifact ) );
        namespace.put( "maven:fullId", Utils.fullArtifactId( artifact ) );
        namespace.put( "maven:groupId", artifact.getGroupId() );
        namespace.put( "maven:groupIdAsDirectory", artifact.getGroupId().replace( '.', '/' ) );
        namespace.put( "maven:artifactId", artifact.getArtifactId() );
        namespace.put( "maven:version", artifact.getBaseVersion() );
        namespace.put( "maven:type", artifact.getProperty( "type", "[void]" ) );
        namespace.put( "maven:extension", artifact.getExtension() );
        if ( artifact.getClassifier() != null )
        {
            namespace.put( "maven:classifier", artifact.getClassifier() );
        }
        else
        {
            namespace.put( "maven:classifier", "[void]" );
        }
        namespace.put( "maven:osgiVersion", Utils.toOsgiVersion( artifact.getBaseVersion(), timestamp ) );
        namespace.put( "maven:buildVersion", Utils.toBuildVersion( artifact.getBaseVersion(), timestamp ) );
        if ( artifact.getFile() != null )
        {
            namespace.put( "maven:fileName", artifact.getFile().getName() );
        }

        return namespace;
    }

    private static InstallableUnit interpolateUnit( final InstallableUnit source, final Map<String, String> namespace )
    {
        final InstallableUnit target = new InstallableUnit();

        String id = source.getId();
        if ( id == null )
        {
            id = "${maven:id}";
        }
        target.setId( interpolate( id, namespace ) );

        String version = source.getVersion();
        if ( version == null )
        {
            version = "${maven:buildVersion}";
        }
        target.setVersion( interpolate( version, namespace ) );

        if ( source.getUpdateDescriptor() != null )
        {
            final UpdateDescriptor interpolated = new UpdateDescriptor();
            interpolated.setId( interpolate( source.getUpdateDescriptor().getId(), namespace ) );
            interpolated.setRange( interpolate( source.getUpdateDescriptor().getRange(), namespace ) );
            interpolated.setSeverity( source.getUpdateDescriptor().getSeverity() );

            target.setUpdateDescriptor( interpolated );
        }

        for ( final Property property : source.getProperties() )
        {
            final Property interpolated = new Property();
            interpolated.setName( interpolate( property.getName(), namespace ) );
            interpolated.setValue( interpolate( property.getValue(), namespace ) );

            target.addProperty( interpolated );
        }

        for ( final ProvidedCapability capability : source.getProvidedCapabilities() )
        {
            final ProvidedCapability interpolated = new ProvidedCapability();
            interpolated.setName( interpolate( capability.getName(), namespace ) );
            interpolated.setNamespace( interpolate( capability.getNamespace(), namespace ) );
            interpolated.setVersion( interpolate( capability.getVersion(), namespace ) );

            target.addProvidedCapability( interpolated );
        }

        for ( final RequiredCapability capability : source.getRequiredCapabilities() )
        {
            target.addRequiredCapability( interpolateRequirement( capability, namespace ) );
        }

        for ( final RequiredCapability capability : source.getMetaRequirements() )
        {
            target.addMetaRequirement( interpolateRequirement( capability, namespace ) );
        }

        for ( final RequiredCapability capability : source.getHostRequirements() )
        {
            target.addHostRequirement( interpolateRequirement( capability, namespace ) );
        }

        for ( final PatchScope scope : source.getPatchScope() )
        {
            final PatchScope interpolated = new PatchScope();
            for ( final RequiredCapability capability : scope.getRequires() )
            {
                interpolated.addRequire( interpolateRequirement( capability, namespace ) );
            }

            target.addPatchScope( interpolated );
        }

        for ( final PatchChange change : source.getPatchChanges() )
        {
            final PatchChange interpolated = new PatchChange();
            if ( change.getFrom() != null && change.getFrom().getRequired() != null )
            {
                final PatchChangeRequirement interpolatedChange = new PatchChangeRequirement();
                interpolatedChange.setRequired( interpolateRequirement( change.getFrom().getRequired(), namespace ) );
                interpolated.setFrom( interpolatedChange );
            }
            if ( change.getTo() != null && change.getTo().getRequired() != null )
            {
                final PatchChangeRequirement interpolatedChange = new PatchChangeRequirement();
                interpolatedChange.setRequired( interpolateRequirement( change.getTo().getRequired(), namespace ) );
                interpolated.setTo( interpolatedChange );
            }

            target.addPatchChange( interpolated );
        }

        if ( source.getPatchLifeCycle() != null )
        {
            target.setPatchLifeCycle( interpolateRequirement( source.getPatchLifeCycle(), namespace ) );
        }

        for ( final InstallableUnitArtifact artifact : source.getArtifacts() )
        {
            final InstallableUnitArtifact interpolated = new InstallableUnitArtifact();
            interpolated.setId( interpolate( artifact.getId(), namespace ) );
            interpolated.setClassifier( interpolate( artifact.getClassifier(), namespace ) );
            interpolated.setVersion( interpolate( artifact.getVersion(), namespace ) );

            target.addArtifact( interpolated );
        }

        if ( source.getTouchpointType() != null )
        {
            final TouchpointType interpolated = new TouchpointType();
            interpolated.setId( interpolate( source.getTouchpointType().getId(), namespace ) );
            interpolated.setVersion( interpolate( source.getTouchpointType().getVersion(), namespace ) );

            target.setTouchpointType( interpolated );
        }

        if ( source.getTouchpointData() != null )
        {
            for ( final TouchpointInstruction sourceInstruction : source.getTouchpointData().getInstructions() )
            {
                if ( target.getTouchpointData() == null )
                {
                    target.setTouchpointData( new TouchpointData() );
                }

                final TouchpointInstruction interpolated = new TouchpointInstruction();
                interpolated.setKey( sourceInstruction.getKey() );
                interpolated.setBody( interpolate( sourceInstruction.getBody(), namespace ) );

                if ( target.getTouchpointData().getInstructions() != null )
                {
                    boolean found = false;
                    for ( final TouchpointInstruction targetInstruction : target.getTouchpointData().getInstructions() )
                    {
                        if ( sourceInstruction.getKey().equals( targetInstruction.getKey() ) )
                        {
                            // TODO does it need an ; between?
                            targetInstruction.setBody( targetInstruction.getBody() + interpolated.getBody() );
                            found = true;
                            break;
                        }
                    }
                    if ( !found )
                    {
                        target.getTouchpointData().addInstruction( interpolated );
                    }
                }
                else
                {
                    target.getTouchpointData().addInstruction( interpolated );
                }
            }
        }

        return target;
    }

    private static RequiredCapability interpolateRequirement( final RequiredCapability capability,
                                                              final Map<String, String> namespace )
    {
        final RequiredCapability interpolated = new RequiredCapability();
        interpolated.setName( interpolate( capability.getName(), namespace ) );
        interpolated.setNamespace( interpolate( capability.getNamespace(), namespace ) );
        interpolated.setRange( interpolate( capability.getRange(), namespace ) );
        interpolated.setFilter( interpolate( capability.getFilter(), namespace ) );
        interpolated.setOptional( capability.isOptional() );
        interpolated.setMultiple( capability.isMultiple() );
        interpolated.setGreedy( capability.isGreedy() );

        interpolated.setMatch( interpolate( capability.getMatch(), namespace ) );
        interpolated.setMatchParameters( interpolate( capability.getMatchParameters(), namespace ) );
        interpolated.setMin( capability.getMin() );
        interpolated.setMax( capability.getMax() );
        return interpolated;
    }

    private InstallableUnit getInstallableUnit( final String id, final String version )
    {
        final IUKey iuKey = new IUKey( id, version );
        InstallableUnit unit = units.get( iuKey );
        if ( unit == null )
        {
            unit = new InstallableUnit();
            unit.setId( id );
            unit.setVersion( version );

            units.put( iuKey, unit );
        }
        return unit;
    }

    private InstallableUnit dependenciesAsRequirements( final DependencyNode node )
    {
        final Map<String, String> namespace = createNamespaceMap( node.getDependency().getArtifact() );

        final InstallableUnit iu = new InstallableUnit();
        iu.setId( interpolate( "${maven:id}", namespace ) );
        iu.setVersion( interpolate( "${maven:buildVersion}", namespace ) );
        for ( final DependencyNode child : node.getChildren() )
        {
            final Map<String, String> childNamespace = createNamespaceMap( child.getDependency().getArtifact() );

            final RequiredCapability req = new RequiredCapability();
            req.setNamespace( "org.sonatype.runtime.installable" );
            req.setName( interpolate( "${maven:id}", childNamespace ) );
            req.setRange( interpolate( String.format( "[%1$s,%1$s]", "${maven:buildVersion}" ), childNamespace ) );

            iu.addRequiredCapability( req );
        }

        return iu;
    }

    private class ArtifactsProcessor
        implements DependencyVisitor
    {

        private final LinkedList<DependencyNode> parents;

        ArtifactsProcessor()
        {
            parents = new LinkedList<DependencyNode>();
        }

        public boolean visitEnter( final DependencyNode node )
        {
            if ( node.getDependency() == null )
            {
                return true;
            }

            for ( final Advice advice : advices )
            {
                final boolean checkSelf = advice.isAppliesToSelf();
                final boolean checkTransitive = advice.isAppliesTransitively();
                final int maxDepth = advice.getAppliesTransitivelyDepth();

                if ( nodeMatches( node, parents, buildFilter( advice ), checkSelf, checkTransitive, 0, maxDepth ) )
                {
                    generateInstallableUnits( advice, node );
                }
            }

            final Artifact artifact = node.getDependency().getArtifact();
            artifacts.add( artifact );
            generateInstallableArtifact( artifact );

            parents.addFirst( node );
            return true;
        }

        public boolean visitLeave( final DependencyNode node )
        {
            if ( node.getDependency() == null )
            {
                return true;
            }
            parents.removeFirst();
            return true;
        }

        private boolean nodeMatches( final DependencyNode node, final LinkedList<DependencyNode> nodeParents,
                                     final DependencyFilter filter, final boolean checkSelf, final boolean checkParent,
                                     final int depth, final int maxDepth )
        {
            DependencyNode parent = null;
            LinkedList<DependencyNode> grandParents;
            if ( !nodeParents.isEmpty() )
            {
                parent = nodeParents.getFirst();
                grandParents = new LinkedList<DependencyNode>( nodeParents );
                grandParents.removeFirst();
            }
            else
            {
                grandParents = new LinkedList<DependencyNode>();
            }

            return ( checkSelf && filter.accept( node, nodeParents ) )
                || ( checkParent && depth < maxDepth && parent != null && nodeMatches( parent, grandParents, filter, true, checkParent, depth + 1, maxDepth ) );
        }

        private DependencyFilter buildFilter( final Advice advice )
        {
            final ArrayList<DependencyFilter> filters = new ArrayList<DependencyFilter>();

            if ( advice.getAppliesTo() != null )
            {
                if ( !advice.getAppliesTo().getIncludes().isEmpty() )
                {
                    filters.add( new PatternInclusionsDependencyFilter( advice.getAppliesTo().getIncludes() ) );
                }
                if ( !advice.getAppliesTo().getExcludes().isEmpty() )
                {
                    filters.add( new PatternExclusionsDependencyFilter( advice.getAppliesTo().getExcludes() ) );
                }
            }

            final AndDependencyFilter artifactFilter = new AndDependencyFilter( filters );
            return artifactFilter;
        }

        private void generateInstallableArtifact( final Artifact artifact )
        {
            final String id = Utils.conflictArtifactId( artifact );
            final String version = Utils.toOsgiVersion( artifact.getBaseVersion(), timestamp );

            final InstallableArtifact installableArtifact = new InstallableArtifact();
            installableArtifact.setId( id );
            installableArtifact.setClassifier( "org.apache.maven.artifact" );
            installableArtifact.setVersion( version );
            installableArtifact.setPath( artifact.getFile().getAbsolutePath() );
            installableArtifact.setRepositoryPath( Utils.pathOf( artifact ) );

            // set maven related properties
            installableArtifact.addProperty( artifactProperty( "org.apache.maven.artifact.groupId", artifact.getGroupId() ) );
            installableArtifact.addProperty( artifactProperty( "org.apache.maven.artifact.artifactId", artifact.getArtifactId() ) );
            installableArtifact.addProperty( artifactProperty( "org.apache.maven.artifact.version", artifact.getVersion() ) );
            installableArtifact.addProperty( artifactProperty( "org.apache.maven.artifact.extension", artifact.getExtension() ) );
            if ( StringUtils.isNotBlank( artifact.getClassifier() ) )
            {
                installableArtifact.addProperty( artifactProperty( "org.apache.maven.artifact.classifier", artifact.getClassifier() ) );
            }

            installableArtifacts.add( installableArtifact );
        }

        private org.sonatype.tycho.p2.publisher.model.artifact.Property artifactProperty( final String name,
                                                                                          final String value )
        {
            final org.sonatype.tycho.p2.publisher.model.artifact.Property property =
                new org.sonatype.tycho.p2.publisher.model.artifact.Property();
            property.setName( name );
            property.setValue( value );
            return property;
        }

        private void generateInstallableUnits( final Advice advice, final DependencyNode node )
        {
            final Artifact artifact = node.getDependency().getArtifact();
            final Map<String, String> namespace = createNamespaceMap( artifact );

            final List<InstallableUnit> ius = new ArrayList<InstallableUnit>();
            final List<InstallableUnit> adviceUnits = advice.getUnits();
            if ( adviceUnits != null )
            {
                ius.addAll( adviceUnits );
            }
            if ( advice.getGenerateSettings() != null )
            {
                final GenerateSettings generateSettings = advice.getGenerateSettings();
                if ( isOn( generateSettings.getP2iu() ) || isOn( generateSettings.getP2manifest() )
                    || isOn( generateSettings.getP2provided() ) || isOn( generateSettings.getP2required() ) )
                {
                    final InstallableUnit[] generatedIUs =
                        bundlePublisher.generate( generateSettings, artifact.getFile() );
                    if ( generatedIUs != null )
                    {
                        ius.addAll( Arrays.asList( generatedIUs ) );
                    }
                }
                if ( isOn( generateSettings.getMavenDependencies() ) )
                {
                    final InstallableUnit generatedIU = dependenciesAsRequirements( node );
                    ius.add( generatedIU );
                }
            }

            for ( final InstallableUnit iu : ius )
            {
                final InstallableUnit interpolated = interpolateUnit( iu, namespace );
                final InstallableUnit unit = getInstallableUnit( interpolated.getId(), interpolated.getVersion() );

                if ( advice.getMetaRequirementSwitch() != null && advice.getMetaRequirementSwitch().isOn() )
                {
                    metaRequirements.add( new IUKey( unit.getId(), unit.getVersion() ) );
                }

                if ( advice.isAppliesToMetaRequirements() )
                {
                    processableUnits.add( interpolated );
                }
                else
                {
                    processableUnitsNotApplicableToMetaRequirements.add( interpolated );
                }
            }
        }

    }

    private static class IUKey
    {
        final String id;

        final String version;

        IUKey( final String id, final String version )
        {
            this.id = id;
            this.version = version;
        }

        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result + ( ( id == null ) ? 0 : id.hashCode() );
            result = prime * result + ( ( version == null ) ? 0 : version.hashCode() );
            return result;
        }

        @Override
        public boolean equals( final Object obj )
        {
            if ( this == obj )
                return true;
            if ( obj == null )
                return false;
            if ( getClass() != obj.getClass() )
                return false;
            final IUKey other = (IUKey) obj;
            if ( id == null )
            {
                if ( other.id != null )
                    return false;
            }
            else if ( !id.equals( other.id ) )
                return false;
            if ( version == null )
            {
                if ( other.version != null )
                    return false;
            }
            else if ( !version.equals( other.version ) )
                return false;
            return true;
        }

    }

    private static boolean isOn( final OnOffSwitch onOffSwitch )
    {
        return onOffSwitch != null && onOffSwitch.isOn();
    }

}